"""Constants module."""

STATUE = "STATUE"
ENCODING = "utf-8"

DEFAULT_HISTORY_SIZE = 30

GENERAL = "general"
COMMANDS = "commands"
CONTEXTS = "contexts"
SOURCES = "sources"

HELP = "help"
ARGS = "args"
CLEAR_ARGS = "clear_args"
ADD_ARGS = "add_args"
ALLOW_LIST = "allow_list"
DENY_LIST = "deny_list"
ALLOWED_CONTEXTS = "allowed_contexts"
DENIED_CONTEXTS = "denied_contexts"
REQUIRED_CONTEXTS = "required_contexts"
ALIASES = "aliases"
PARENT = "parent"
ALLOWED_BY_DEFAULT = "allowed_by_default"
VERSION = "version"
MODE = "mode"
HISTORY_SIZE = "history_size"
CACHE = "cache"

TEMPLATE_NAME_REGEX = r"^[A-Za-z]\w*$"
DATETIME_FORMAT = "%m/%d/%Y, %H:%M:%S"
BAR_FORMAT = "{l_bar}{bar}| {n_fmt}/{total_fmt}"
MAIN_BAR_COLOR = "blue"
SECONDARY_BAR_COLOR = "yellow"
